from pathlib import Path
from itertools import chain

from ..tools import dumps
from ..tools import loads

from ..elements import Edge
from ..elements import Vertex

from ..managers import EdgeManager
from ..managers import VertexManager

from ..packing import pack
from ..packing import unpack
from ..tools import LazySliceGenerator
from ..exceptions import ElementNotFound


class Transaction:

    def __init__(self, storage):
        self._txn = None
        self.vertex = VertexManager(self)
        self.edge = EdgeManager(self)
        self._storage = storage
        self._dirty = set()

    def commit(self):
        for element in self._dirty:
            self._storage._save(element)
        # self._txn.commit()

    def rollback(self):
        # self._txn.abort()
        pass


class FileStore:

    class Cursor():

        def __init__(self, store):
            self._store = store
        
        def first(self):
            self._iter = map(lambda x: x.name, self._store._path.iterdir())
            return self.next()

        def set_range(self, prefix):
            prefix = prefix.replace('\x00', ':')
            iterator = self._store._path.iterdir()
            names = sorted(map(lambda x: x.name, iterator))
            for index, name in enumerate(names):
                if name.startswith(prefix):
                    break
            self._iter = iter(names[index:])
            return self.next()

        def next(self):
            try:
                name = next(self._iter)
            except StopIteration:
                return None
            else:
                name = name.replace('\x00', ':')
                return name, self._store.get(name)   
            
    def cursor(self, txn):
        return self.Cursor(self)
    
    def __init__(self, path):
        self._path = path
        if not self._path.exists():
            self._path.mkdir(parents=True)

    def put(self, key, value, txn=None):
        # packing function adds NULL char to "reset the search"
        # NULL char can be used on the filesystem so we replace
        # we two dots which are not supposed to appear in property
        # names
        key = key.replace('\00', ':')
        path = self._path / key
        with path.open('wb') as f:
            f.write(value)

    def get(self, key, txn=None):
        # packing function adds NULL char to "reset the search"
        # NULL char can be used on the filesystem so we replace
        # we two dots which are not supposed to appear in property
        # names
        key = key.replace('\00', ':')
        path = self._path / key
        try:
            with path.open('rb') as f:
                return f.read()
        except FileNotFoundError:
            return None

    def delete(self, key, txn):
        path = self._path / key
        path.unlink()


class FileStorage:
    """File based storage"""

    def __init__(self, path):
        self._path = Path(path).absolute()
        # create directory (and its parents) if it doesn't exists
        if not self._path.exists():
            self._path.mkdir(parents=True)

        # vertices
        self._vertices = FileStore(self._path / 'vertices')
        self._vertices_labels = FileStore(self._path / 'vertices_labels')
        # edges
        self._edges = FileStore(self._path / 'edges')
        self._edges_labels = FileStore(self._path / 'edges_labels')

    def _txn(self):
        return Transaction(self)

    def close(self):
        # self._vertices.close()
        # self._edges.close()
        # self._env.close()
        pass

    def _save(self, element):
        txn = element._manager._txn._txn
        if isinstance(element, Vertex):
            value = (
                element._label,
                element._outgoings,
                element._incomings,
                element.properties.freeze()
            )
            store = self._vertices
            index = pack('SS', element._label, element._identifier)
            self._vertices_labels.put(index, b'', txn)
        else:
            value = (
                element._start,
                element._label,
                element._end,
                element.properties.freeze()
            )
            store = self._edges
            index = pack('SS', element._label, element._identifier)
            self._edges_labels.put(index, b'', txn)
        key = element._identifier
        value = dumps(value)
        store.put(key, value, txn)

    def lookup(self, txn, store_name, fmt, *lookup):
        """Look up for entries based on `lookup`.

        The function only support string and integers. This must take into
        account how the queried database is constructed (see below). Querying
        happens over the key and returns the identifiers.

        Label indexes ``'vertices.labels'`` and ``'edges.labels'`` expects
        only a label name so ``lookup`` equals to ``('some_label',)``.

        Properties indexes ``'vertices.properties'`` and ``'edges.properties'``
        expects an ``attribute_name`` and an ``attribute_value`` which can be
        a ``str`` or an ``int``.

        This can also be used to check for the existence of a entry, for
        instance, it's a quick way given the tuple (name, value, identifier)
        to know if the element associated with ``identifier`` has property
        ``name`` associated with ``value``. It's **probably** quickest than
        lloading the object.

        .. warning:: Don't do funky things with ``*key`` only what is
                     documented is supported (and it's for internal use only)!
        """
        store = getattr(self, store_name)
        cursor = store.cursor(txn)
        index = pack(fmt[:len(lookup)], *lookup)
        next = cursor.set_range(index)

        while next:
            values, _ = next
            values = values.replace(':', '\x00').encode('utf8')
            values = unpack(fmt, values)
            # check that it's a good candidate
            if all(map(lambda x: x[0] == x[1], zip(lookup, values))):
                yield values
                next = cursor.next()
            else:
                break

    # FIXME: same as bsddb3
    def _delete(self, element):
        key = element._identifier
        txn = element._manager._txn._txn
        if isinstance(element, Vertex):
            for edge in chain(element.outgoings(), element.incomings()):
                edge.delete()
            # delete this element from the store
            self._vertices.delete(key, txn)
        else:
            start = element.start()
            start._outgoings[element._label].remove(element._identifier)
            start._dirty()
            end = element.end()
            end._incomings[element._label].remove(element._identifier)
            end._dirty()
            # delete this element from the store
            self._edges.delete(key, txn)

    def _element(self, storage, ElementClass, identifier, manager, txn=None):
        try:
            element = manager._cache[identifier]
        except KeyError:
            key = identifier
            element = storage.get(key, txn=manager._txn._txn)
            if not element:
                msg = 'Element with id=%s not found' % identifier
                raise ElementNotFound(msg)
            element = loads(element)
            element = ElementClass(manager, identifier, *element)
            manager._cache[identifier] = element
        return element

    get_vertex = lambda s, *a, **kw: s._element(s._vertices, Vertex, *a, **kw)
    get_edge = lambda s, *a, **kw: s._element(s._edges, Edge, *a, **kw)

    def _slice(self, storage, ElementClass, txn):
        def generator():
            cursor = storage.cursor(txn=txn._txn)
            next = cursor.first()
            while next:
                yield next
                next = cursor.next()

        def wrapper(pair):
            identifier, element = pair
            element = loads(element)
            if ElementClass is Vertex:
                return ElementClass(txn.vertex, identifier, *element)
            else:
                return ElementClass(txn.edge, identifier, *element)

        return LazySliceGenerator(wrapper, generator())

    slice_vertices = lambda s, *a, **kw: s._slice(s._vertices, Vertex, *a, **kw)
    slice_edges = lambda s, *a, **kw: s._slice(s._edges, Edge, *a, **kw)

    # this is the same code as bsddb3
    def create_vertex(self, label, txn):
        """Called by a vertex transaction manager to build a vertex.

        For wrapping a Vertex already in the database one must use
        the constructor."""
        identifier = txn.vertex._random_identifier()
        vertex = Vertex(txn.vertex, identifier, label, dict(), dict(), dict())
        txn.vertex._cache[identifier] = vertex
        vertex._dirty()
        return vertex

    def create_edge(self, start, label, end, txn):
        """Called by a edge transaction manager to build an edge.

        For wrapping a Vertex already in the database one must use
        the constructor."""
        identifier = txn.edge._random_identifier()
        edge = Edge(
            txn.edge,
            identifier,
            start._identifier,
            label,
            end._identifier,
            dict()
        )
        txn.edge._cache[identifier] = edge
        edge._dirty()
        # update start and end
        try:
            start._outgoings[edge._label].append(edge.identifier())
        except KeyError:
            start._outgoings[edge._label] = [edge.identifier()]
        try:
            end._incomings[edge._label].append(edge.identifier())
        except KeyError:
            end._incomings[edge._label] = [edge.identifier()]
        # mark vertices as dirty to be updated
        start._dirty()
        end._dirty()
        return edge
