#!/usr/bin/env python3
from shutil import rmtree
from unittest import TestCase

from ajgu import GraphDatabase
from ajgu.storage.bsddb3 import BSDDB3

from ajgu.tools import dumps
from ajgu.tools import loads

from ajgu.uzelmumu import run
from ajgu.uzelmumu import graph


class DatabaseTestCase(TestCase):

    def tearDown(self):
        try:
            rmtree('/tmp/ajgu')
        except FileNotFoundError:
            pass


class TestGraphDatabase(DatabaseTestCase):

    storage_class = BSDDB3

    def test_create(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        graph.close()

    def test_create_vertex(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            v = txn.vertex.create('new')
        self.assertTrue(v)

    def test_create_and_get_vertex(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            v = txn.vertex.create('get-vertex')
        with graph.transaction() as txn:
            v = txn.vertex.get(v.identifier())
        self.assertTrue(v)

    def test_create_modify_and_get_vertex(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            v = txn.vertex.create('create-modify-and-get-vertex')
            v.properties['value'] = 'key'
        with graph.transaction() as txn:
            v = txn.vertex.get(v.identifier())
        self.assertTrue(v.properties['value'] == 'key')

    def test_create_and_get_modify_and_get_again_vertex(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            v = txn.vertex.create('create-and-get-modify-and-get-again-vertex')
        with graph.transaction() as txn:
            v = txn.vertex.get(v.identifier())
            v.properties['value'] = 'key'
        with graph.transaction() as txn:
            v = txn.vertex.get(v.identifier())
        self.assertTrue(v.properties['value'] == 'key')

    def test_test_create_modify_and_get_edge(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            start = txn.vertex.create('start')
            end = txn.vertex.create('end')
            edge = txn.edge.create(start, 'héllo', end)
            edge.properties['who'] = 'world'
        with graph.transaction() as txn:
            edge = txn.edge.get(edge.identifier())
        self.assertTrue(edge.properties['who'] == 'world')

    def test_create_edge_and_checkout_vertex(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            start = txn.vertex.create('start')
            end = txn.vertex.create('end')
            txn.edge.create(start, 'héllo', end)
        # retrieve start and end
        with graph.transaction() as txn:
            start = txn.vertex.get(start.identifier())
            end = txn.vertex.get(end.identifier())
            self.assertTrue(list(start.outgoings()) and list(end.incomings()))

    def test_create_edge_and_filtered_edges(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            # vertices
            start = txn.vertex.create('start')
            end = txn.vertex.create('end')
            # first edge
            txn.edge.create(start, 'héllo', end)
            # second edge
            txn.edge.create(start, 'world', end)
        # retrieve start and end
        with graph.transaction() as txn:
            start = txn.vertex.get(start.identifier())
            end = txn.vertex.get(end.identifier())
        assert len(list(start.outgoings())) == 2
        assert len(list(end.incomings('héllo'))) == 1
        assert len(list(end.incomings('world'))) == 1
        assert len(list(start.outgoings('héllo'))) == 1
        assert len(list(start.outgoings('world'))) == 1

    def test_delete_node(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            v = txn.vertex.create('delete-node')
            v.properties['value'] = 'key'
        # retrieve it
        with graph.transaction() as txn:
            v = txn.vertex.create(v.identifier())
        if not v:
            return False
        # delete it
        with graph.transaction() as txn:
            v.delete()
        with graph.transaction() as txn:
            try:
                v = txn.vertex.get(v.identifier())
            except:
                v = None
            self.assertIsNone(v)

    def test_delete_edge(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            # vertices
            start = txn.vertex.create('start')
            end = txn.vertex.create('end')
            # first edge
            edge = txn.edge.create(start, 'héllo', end)

        # delete edge
        with graph.transaction() as txn:
            edge = txn.edge.get(edge.identifier())
            edge.delete()

        # check end and start node were updated
        with graph.transaction() as txn:
            start = txn.vertex.get(start.identifier())
            self.assertEquals(len(list(start.outgoings())), 0)
            end = txn.vertex.get(end.identifier())
            self.assertEquals(len(list(end.incomings())), 0)

    def test_delete_one_edge(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            # vertices
            start = txn.vertex.create('start')
            end = txn.vertex.create('end')
            # first edge
            edge = txn.edge.create(start, 'héllo', end)
            edge = txn.edge.create(end, 'héllo', start)

        with graph.transaction() as txn:
            start = txn.vertex.get(start.identifier())
            self.assertEquals(len(list(start.outgoings())), 1)
            self.assertEquals(len(list(start.incomings())), 1)
            end = txn.vertex.get(end.identifier())
            self.assertEquals(len(list(end.incomings())), 1)
            self.assertEquals(len(list(end.outgoings())), 1)

        # delete edge
        with graph.transaction() as txn:
            edge = txn.edge.get(edge.identifier())
            edge.delete()

        # check end and start node were updated
        with graph.transaction() as txn:
            start = txn.vertex.get(start.identifier())
            self.assertEquals(len(list(start.outgoings())), 1)
            end = txn.vertex.get(end.identifier())
            self.assertEquals(len(list(end.incomings())), 1)

    def create_feeds(self):
        database = GraphDatabase(self.storage_class('/tmp/ajgu'))
        # create user
        with database.transaction() as txn:
            amirouche = txn.vertex.create('user')
            amirouche.properties['name'] = 'amirouche'

        # create feed
        with database.transaction() as txn:
            amirouche = txn.vertex.get(amirouche.identifier())

            hypermove = txn.vertex.create('feed')
            hypermove.properties['url'] = 'http://hyperdev.fr/atom.xml'
            txn.edge.create(amirouche, 'feeds', hypermove)

            hypermove = txn.vertex.create('feed')
            url = 'http://www.hyperdev.fr/keyword/fran%C3%A7ais/atom.xml'
            hypermove.properties['url'] = url
            txn.edge.create(amirouche, 'feeds', hypermove)

        with database.transaction() as txn:
            amirouche = txn.vertex.get(amirouche.identifier())
            for feed in map(lambda x: x.end(), amirouche.outgoings('feeds')):
                from requests import get
                from lxml.etree import fromstring
                xml = fromstring(get(feed.url).text)
                ns = dict(a='http://www.w3.org/2005/Atom')
                list(map(
                    print,
                    xml.xpath('/a:feed/a:entry/a:title/text()', namespaces=ns)
                ))


class TestVertexManager(DatabaseTestCase):

    storage_class = BSDDB3

    def test_all_vertices(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            txn.vertex.create('one')
            txn.vertex.create('two')
            txn.vertex.create('three')

        with graph.transaction() as txn:
            labels = list(map(lambda x: x.label(), txn.vertex.all()))
            self.assertEquals(set(labels), set(['one', 'two', 'three']))

    def test_slice_vertices(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            txn.vertex.create('one')
            txn.vertex.create('two')
            txn.vertex.create('three')

        with graph.transaction() as txn:
            count = len(list(txn.vertex.slice(1, 3)))
            self.assertEquals(2, count)

    def test_slice_count(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            txn.vertex.create('one')
            txn.vertex.create('two')
            txn.vertex.create('three')

        with graph.transaction() as txn:
            count = txn.vertex.count()
            self.assertEquals(3, count)

    def test_label(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            vertex = txn.vertex.create('one')
            vertex.props['name'] = 'one one'
            vertex = txn.vertex.create('one')
            vertex.props['name'] = 'one two'
            txn.vertex.create('three')

        with graph.transaction() as txn:
            count = txn.vertex.label('one').count()
            self.assertEquals(2, count)
            ones = txn.vertex.label('one').all()
            names = set(map(lambda x: x.props.name, ones))
            self.assertEquals({'one one', 'one two'}, names)


class TestEdgeManagerMethods(DatabaseTestCase):

    storage_class = BSDDB3

    def test_all_vertices(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            txn.edge.create(
                txn.vertex.create('start'),
                'one',
                txn.vertex.create('end')
            )
            txn.edge.create(
                txn.vertex.create('start'),
                'two',
                txn.vertex.create('end')
            )
            txn.edge.create(
                txn.vertex.create('start'),
                'three',
                txn.vertex.create('end')
            )

        with graph.transaction() as txn:
            labels = list(map(lambda x: x.label(), txn.edge.all()))
            self.assertEquals(set(labels), set(['one', 'two', 'three']))

    def test_slice_vertices(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            txn.edge.create(
                txn.vertex.create('start'),
                'one',
                txn.vertex.create('end')
            )
            txn.edge.create(
                txn.vertex.create('start'),
                'two',
                txn.vertex.create('end')
            )
            txn.edge.create(
                txn.vertex.create('start'),
                'three',
                txn.vertex.create('end')
            )

        with graph.transaction() as txn:
            count = len(list(txn.edge.slice(1, 3)))
            self.assertEquals(2, count)

    def test_slice_count(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            txn.edge.create(
                txn.vertex.create('start'),
                'one',
                txn.vertex.create('end')
            )
            txn.edge.create(
                txn.vertex.create('start'),
                'two',
                txn.vertex.create('end')
            )
            txn.edge.create(
                txn.vertex.create('start'),
                'three',
                txn.vertex.create('end')
            )

        with graph.transaction() as txn:
            count = txn.edge.count()
            self.assertEquals(3, count)

    def test_label(self):
        graph = GraphDatabase(self.storage_class('/tmp/ajgu'))
        with graph.transaction() as txn:
            one = txn.vertex.create('vertex')
            two = txn.vertex.create('vertex')
            txn.edge.create(one, 'one', two)
            txn.edge.create(one, 'one', two)
            txn.edge.create(one, 'two', two)

        with graph.transaction() as txn:
            count = txn.edge.label('one').count()
            self.assertEquals(2, count)


class TestUzelmumuQuery(DatabaseTestCase):

    storage_class = BSDDB3

    def test_run_uzelmumu_query(self):
        # populate database
        database = GraphDatabase(self.storage_class('/tmp/ajgu'))

        # create user
        with database.transaction() as txn:
            amirouche = txn.vertex.create('user')
            amirouche.properties['name'] = 'amirouche'

        # create feed
        with database.transaction() as txn:
            amirouche = txn.vertex.get(amirouche.identifier())

            hypermove = txn.vertex.create('feed')
            hypermove.properties['url'] = 'http://hyperdev.fr/atom.xml'
            txn.edge.create(amirouche, 'feeds', hypermove)

            hypermove = txn.vertex.create('feed')
            url = 'http://www.hyperdev.fr/keyword/fran%C3%A7ais/atom.xml'
            hypermove.properties['url'] = url
            txn.edge.create(amirouche, 'feeds', hypermove)

        query = graph().vertex.get(amirouche.identifier()).outgoings('feeds')
        query = query.end().freeze()
        query = query.serialize()
        query = loads(dumps(query))

        feeds = run(database, query)

        urls = list()
        for feed in feeds:
            urls.append(feed['properties']['url'])
        if not (urls[0] == 'http://hyperdev.fr/atom.xml'):
            return False
        url = 'http://www.hyperdev.fr/keyword/fran%C3%A7ais/atom.xml'
        if not (urls[1] == url):
            return False
        return True
